from typing import Any, List, Dict, Text, Optional, Tuple, TypeVar, Union, Awaitable
import logging
import redis
from ewoxdbredis.repositories.redis_base_repository import RedisBaseRepository

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class RedisSortedSetRepository(RedisBaseRepository):
    def __init__(self, table_name:str="") -> None:
        super().__init__(table_name)


    async def get(self, connection:T, key:str) -> float:
        """ Get the score of a key in the sorted set. If the key does not exist, return -1. """
        if (self._table_name):
            index:int = await connection.zrevrank(self._table_name, key)
            if (index != -1):
                items_inner = await connection.zrange(self._table_name, index, index, desc=True, withscores=True)
                if (items_inner is not None):
                    tup:Tuple[str, float] = items_inner[0]
                    val:float = tup[1]
                    return val

        return -1


    async def delete(self, connection:T, key:str) -> bool:
        """ Delete a key from the sorted set. Returns True if the key was deleted, False otherwise. """
        if (self._table_name):
            await connection.zrem(self._table_name, key)
            return True

        return False


    async def get_num(self, connection:T) -> int:
        """ Get the number of items in the sorted set. Returns 0 if the table name is not set. """
        num:int = 0
        if (self._table_name):
            num = await connection.zcard(self._table_name)
        
        return num


    async def set(self, connection:T, key:str, score:float) -> bool:
        """ Returns true if the order of the set has changed """
        if (self._table_name):
            dict_score:Dict = dict()
            dict_score[key] = score
            val = await connection.zadd(self._table_name, dict_score, ch=True)
            if (val == 1):
                return True
        
        return False


    async def set_items(self, connection:T, items:Dict) -> bool:
        """ Set multiple items in the sorted set. Returns True if the items were added successfully. """
        if (self._table_name):
            await connection.zadd(self._table_name, items)
            return True
        
        return False


    async def get_items(self, connection:T, skip:int, num:int) -> List[Tuple[str, float]]:
        """ Get items from the sorted set with pagination. Returns a list of tuples (key, score). """
        if (self._table_name):
            items:List[Tuple[str, float]] = []

            items_inner = await connection.zrange(self._table_name, skip, num, desc=True, withscores=True)
            if (items_inner is not None):
                for tup in items_inner:
                    key:str = tup[0].decode()
                    val:float = tup[1]
                    items.append((key, val))

                return items

        return []


    async def increment_by(self, connection:T, key:str, value:float) -> bool:
        """ Increment the score of a key in the sorted set by a specified value. Returns True if the key was found and incremented. """
        if (self._table_name):
            val = await connection.zincrby(self._table_name, value, key)
            return True

        return False


    async def get_rank(self, connection:T, key:str) -> int:
        """ The rank (or index) is 0-based, which means that the member with the lowest score has rank 0. """
        index:int = -1
        if (self._table_name):
            index = await connection.zrank(self._table_name, key)

        return index


    async def get_range(self, connection:T, start:int, end:int, descending:bool=False) -> List[tuple[Any, float]]:
        """ Get a range of items from the sorted set. If descending is True, get items in descending order. """
        if (self._table_name):
            return await connection.zrange(self._table_name, start, end, descending, withscores=True)

        return []


    async def get_range_by_score(self, connection:T, min_score:float, max_score:float) -> List[tuple[Any, float]]:
        """ Get items from the sorted set within a score range. """
        if (self._table_name):
            return await connection.zrangebyscore(self._table_name, min_score, max_score, withscores=True)

        return []


if __name__ == "__main__":
    from ewoxdbredis.connection.db_connection import DBConnection
    rep = RedisSortedSetRepository()
    rep.set_table_name("testSortedSet")
    with DBConnection() as connection:
        num:int = rep.get_num(connection)
        rep.set(connection, "123", 100)
        rep.set(connection, "124", 999)
        rep.set(connection, "125", 101)
        rep.set(connection, "126", 9)
        val:float = rep.get(connection, "123")
        val:float = rep.get(connection, "126")
        rep.increment_by(connection, "126", 1)
        val:float = rep.get(connection, "126")
        print(val)

        val:float = rep.get(connection, "123")
        items = rep.get_items(connection, 1, 300)
        print(items)
        rep.delete(connection, "126")
        items = rep.get_items(connection, 1, 3)
        print(items)



from typing import Any, List, Dict, Text, Optional, Tuple, TypeVar, Union, Awaitable
import os
import time
import logging
import redis
from redis.backoff import ExponentialBackoff
from redis.retry import Retry
from ewoxcore.constants.server_env import ServerEnv
from ewoxdbredis.clients.iredis_client import IRedisClient
from ewoxdbredis.settings.connection_settings import ConnectionSettings

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class RedisClusterClient(IRedisClient):
    def __init__(self, settings:ConnectionSettings=ConnectionSettings()) -> None:
        self._settings:ConnectionSettings = settings
        self._connection:T = None


    def get_connection(self) -> T:
        """ Get the current connection. """
        if (self._connection is None):
            self._connection = self._connect()
        return self._connection
    

    async def setup(self) -> None:
        """ Setup the connection if it is not already established. """
        try:
            if (self._connection is None):
                self._connection = self.get_connection()
                if (self._connection):
                    ok:bool = await self.is_up()
                    if not ok:
                        raise ConnectionError("Redis PING returned falsy result")
                    print(f"Connected to Redis instance at {self._settings.host}:{self._settings.port}")
        except Exception as error:
            print(f"RedisClusterClient.setup. Error: {str(error)}")
            logging.error(f"RedisClusterClient.setup. Error: {error}")


    async def dispose(self) -> None:
        """ Dispose the current connection. """
        if (self._connection is not None):
            try:
                await self._connection.close()
            except Exception as error:
                logging.error(f"RedisClusterClient.dispose. Error: {error}")
            finally:
                self._connection = None


    def _connect(self) -> Optional[redis.asyncio.RedisCluster]:
        rc:redis.asyncio.RedisCluster = None

        try:
            rc = redis.asyncio.RedisCluster(host=self._settings.host, port=self._settings.port, 
                                    decode_responses=self._settings.decode_responses,
                                    ssl=self._settings.use_ssl,
                                    retry=Retry(ExponentialBackoff(), retries=5))
        except Exception as error:
            logging.error(f"RedisClusterClient.connect - Error: {error}")

        return rc


    async def is_up(self) -> bool:
        try:
            return await self._connection.ping()
        except Exception:
            return False


from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable

class PublisherSettings():
    def __init__(self,
                 environment:str,
                 stream_name:str
                ) -> None:
        self.stream_name:str = f"{environment}:{stream_name}"



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable
import os
from ewoxcore.constants.server_env import ServerEnv


class ConnectionSettings():
    def __init__(self,
                 env_host:str="REDIS_HOST",
                 env_port:str="REDIS_PORT",
                 use_ssl:bool=False,
                 is_cluster:bool=False,
                 decode_responses:bool=True) -> None:
        self.host:str = os.getenv(env_host)
        self.port:int = 6379
        if (os.getenv(env_port)):
            self.port = int(os.getenv(env_port))

        self.use_ssl:bool = use_ssl
        self.is_cluster:bool = is_cluster
        self.decode_responses:bool = decode_responses



from typing import Any, List, Dict, Mapping, Text, Optional, Tuple, TypeVar, Union, Awaitable
from datetime import date, datetime, timedelta
import logging
import redis
from ewoxdbredis.repositories.redis_base_repository import RedisBaseRepository
from ewoxdbredis.settings.connection_settings import ConnectionSettings

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class RedisExpireRepository(RedisBaseRepository):
    def __init__(self, table_name:str="") -> None:
        super().__init__(table_name)


    def _get_internal_key(self, key:str) -> str:
        return f"{self._table_name}_{key}"


    async def set_expire(self, connection:T, key:str, value:str, expire:timedelta) -> bool:
        """ Add a key/value that expires in x-time from now. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            await connection.set(internal_key, value)
            expire_at:datetime = datetime.now() + expire
            await connection.expireat(internal_key, expire_at)
            return True

        return False


    async def set_expire_at(self, connection:T, key:str, value:str, expire_at:datetime) -> bool:
        """ Add a key/value that expires at. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            await connection.set(internal_key, value)
            await connection.expireat(internal_key, expire_at)
            return True
        
        return False


    async def set(self, connection:T, key:str, value:str) -> bool:
        """ Set a key/value without expiration. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            await connection.set(internal_key, value)
            return True
        
        return False


    async def get(self, connection:T, key:str) -> str:
        """ Get a value by its key. If the key does not exist, return an empty string. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            value:str = await connection.get(internal_key)
            return value.decode() if (value is not None) else ""
        
        return ""


    async def delete(self, connection:T, key:str) -> bool:
        """ Delete a key from the Redis store. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            await connection.delete(internal_key)
            return True
        
        return False


    async def exist(self, connection:T, key:str) -> bool:
        """ Check if a key exists in the Redis store. """
        if (self._table_name):
            internal_key:str = self._get_internal_key(key)
            return await connection.exists(internal_key)

        return False



if __name__ == "__main__":
    from ewoxdbredis.connection.db_connection import DBConnection

    rep = RedisExpireRepository()
    rep.set_table_name("cache")
    with DBConnection() as connection:
        rep.set_expire(connection, "cacheKey1", "123", timedelta(seconds=20))




from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable

class ConsumerSettings():
    def __init__(self,
                 environment:str,
                 group_name:str,
                 stream_name:str,
                 consumer_name:str,
                 deleteAfterAck:bool=True,
                 claim_after_seconds:int=60*30,  # 30 minutes
                 count:int=10,
                 max_inflight:int=20
                ) -> None:
        self.group_name:str = f"{environment}:{group_name}"
        self.stream_name:str = f"{environment}:{stream_name}"
        self.consumer_name:str = consumer_name
        self.deleteAfterAck:bool = deleteAfterAck
        self.claim_after_seconds:int = claim_after_seconds
        self.count:int = count
        self.max_inflight:int = max_inflight


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type, cast
import redis
import asyncio
from redis.exceptions import ConnectionError, TimeoutError
from tenacity import retry, wait_exponential, wait_fixed, stop_after_attempt, RetryError
from ewoxdbredis.clients.iredis_client import IRedisClient
from ewoxdbredis.settings.consumer_settings import ConsumerSettings
from ewoxcore.message.imessage_consumer import IMessageConsumer
from ewoxcore.message.message_args import MessageArgs

C = TypeVar("C", redis.RedisCluster, redis.Redis)


class MessageStreamConsumer():
    def __init__(self, settings:ConsumerSettings, client:IRedisClient, consumer:IMessageConsumer) -> None:
        self._settings:ConsumerSettings = settings
        self._client:IRedisClient = client
        self._consumer:IMessageConsumer = consumer


    async def subscribe(self) -> None:
        """ Subscribe to the message stream. """
        connection:C = self._client.get_connection()
        if not connection:
            raise Exception("No Redis connection available.")

        await self._create_group(connection)

        while True:
            try:
                await self._process(connection)
            # except RetryError as re:
            #     print(f"Error: Sending to dead-letter queue.")
            #     connection.xadd(DEAD_STREAM, data)
            #     connection.xack(self._settings.stream_name, self._settings.group_name, mid)
            except Exception as e:
                print("[!] Unexpected error:", e)
                await asyncio.sleep(2)


    async def _create_group(self, connection:C) -> None:
        try:
            connection.xgroup_create(name=self._settings.stream_name, groupname=self._settings.group_name, id="0", mkstream=True)
            print(f"Created consumer group '{self._settings.group_name}' on stream '{self._settings.stream_name}'")
        except redis.ResponseError as e:
            if "BUSYGROUP" in str(e):
                print(f"Consumer group '{self._settings.group_name}' already exists.")
            else:
                raise


    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2))
    async def process_message(self, msg_id:str, msg_data:Dict[str, Any]) -> None:
        print(f"Processing message {msg_id}: {str(msg_data)}")
        # args:MessageArgs = MessageArgs(
        #     data=msg_data.get("data", ""),
        #     correlation_id=msg_data.get("correlationId", ""),
        #     service_name=msg_data.get("serviceName", ""),
        #     server_name=msg_data.get("serverName", ""),
        #     send_at=msg_data.get("sendAt", None),
        # )

        await self._consumer.on_consume(
            correlation_id=msg_data.get("correlationId", ""),
            json64=msg_data.get("data", ""),
            service_name=msg_data.get("serviceName", ""))


    @retry(
        wait=wait_exponential(min=1, max=10),
        stop=stop_after_attempt(5),
        retry_error_callback=lambda retry_state: None,
        retry=lambda retry_state: isinstance(
            retry_state.outcome.exception(),
            (ConnectionError, TimeoutError, OSError),
        )
    )    
    async def _process(self, connection:C) -> None:
        results = connection.xreadgroup(
            groupname=self._settings.group_name,
            consumername=self._settings.consumer_name,
            streams={self._settings.stream_name: '>'},
            count=10,
            block=5000
        )

        if results:
            for stream, messages in results:
                for msg_id, msg_data in messages:
                    try:
                        await self.process_message(msg_id, msg_data)
                        self._ack_message(connection, msg_id)
                    except Exception as e:
                        print(f"Failed to process message {msg_id}: {e}")
                        
                        # Write to dead letter stream
                        connection.xadd(
                            f"{self._settings.stream_name}:deadletter",
                            fields={
                                "original_id": msg_id,
                                "data": str(msg_data),
                                "error": str(e),
                            },
                            maxlen=10000,  # keep only latest 10,000 entries
                            approximate=True
                        )                        
                        # Ack the original message
                        self._ack_message(connection, msg_id)
        else:
            # If no messages were processed, check for pending messages to claim
            await self._claim_pending(connection)


    def _ack_message(self, connection:C, msg_id:str) -> None:
        """ Acknowledge a message in the stream. """
        connection.xack(self._settings.stream_name, self._settings.group_name, msg_id)
        if (self._settings.deleteAfterAck):
            connection.xdel(self._settings.stream_name, msg_id)


    async def _claim_pending(self, connection:C) -> None:
        """ Claim pending messages from failed consumers older than X seconds. """
        pending = connection.xpending_range(self._settings.stream_name, self._settings.group_name, min="-", max="+", count=10)
        for p in pending:
            message_id, consumer, elapsed_ms, delivery_count = p
            claim_after_ms:int = self._settings.claim_after_seconds * 1000
            if elapsed_ms > claim_after_ms:
                print(f"Claiming message {message_id} from {consumer}")
                claimed = connection.xclaim(
                    name=self._settings.stream_name,
                    groupname=self._settings.group_name,
                    consumername=self._settings.consumer_name,
                    min_idle_time=claim_after_ms,
                    message_ids=[message_id],
                )

                for mid, data in claimed:
                    await self.process_message(mid, data)
                    connection.xack(self._settings.stream_name, self._settings.group_name, mid)
